@@ -6,7 +6,7 @@
 its pattern matching. On a Unix or Win32 system it can recurse into
 directories.
 
-           Copyright (c) 1997-2010 University of Cambridge
+           Copyright (c) 1997-2011 University of Cambridge
 
 -----------------------------------------------------------------------------
 Redistribution and use in source and binary forms, with or without
@@ -74,9 +74,9 @@ typedef int BOOL;
 #define OFFSET_SIZE 99
 
 #if BUFSIZ > 8192
-#define MBUFTHIRD BUFSIZ
+#define PATBUFSIZE BUFSIZ
 #else
-#define MBUFTHIRD 8192
+#define PATBUFSIZE 8192
 #endif
 
 /* Values for the "filenames" variable, which specifies options for file name
@@ -135,6 +135,7 @@ static char *colour_string = (char *)"1;31";
 static char *colour_option = NULL;
 static char *dee_option = NULL;
 static char *DEE_option = NULL;
+static char *main_buffer = NULL;
 static char *newline = NULL;
 static char *pattern_filename = NULL;
 static char *stdin_name = (char *)"(standard input)";
@@ -159,6 +160,8 @@ static pcre *exclude_dir_compiled = NULL;
 static int after_context = 0;
 static int before_context = 0;
 static int both_context = 0;
+static int bufthird = PCREGREP_BUFSIZE;
+static int bufsize = 3*PCREGREP_BUFSIZE;
 static int dee_action = dee_READ;
 static int DEE_action = DEE_READ;
 static int error_count = 0;
@@ -166,6 +169,12 @@ static int filenames = FN_DEFAULT;
 static int only_matching = -1;
 static int process_options = 0;
 
+#ifdef SUPPORT_PCREGREP_JIT
+static int study_options = PCRE_STUDY_JIT_COMPILE;
+#else
+static int study_options = 0;
+#endif
+
 static unsigned long int match_limit = 0;
 static unsigned long int match_limit_recursion = 0;
 
@@ -186,8 +195,8 @@ static BOOL utf8 = FALSE;
 
 /* Structure for options and list of them */
 
-enum { OP_NODATA, OP_STRING, OP_OP_STRING, OP_NUMBER, OP_OP_NUMBER,
-       OP_PATLIST };
+enum { OP_NODATA, OP_STRING, OP_OP_STRING, OP_NUMBER, OP_LONGNUMBER,
+       OP_OP_NUMBER, OP_PATLIST };
 
 typedef struct option_item {
   int type;
@@ -214,43 +223,51 @@ used to identify them. */
 #define N_LBUFFER      (-12)
 #define N_M_LIMIT      (-13)
 #define N_M_LIMIT_REC  (-14)
+#define N_BUFSIZE      (-15)
+#define N_NOJIT        (-16)
 
 static option_item optionlist[] = {
-  { OP_NODATA,    N_NULL,   NULL,              "",              "  terminate options" },
-  { OP_NODATA,    N_HELP,   NULL,              "help",          "display this help and exit" },
-  { OP_NUMBER,    'A',      &after_context,    "after-context=number", "set number of following context lines" },
-  { OP_NUMBER,    'B',      &before_context,   "before-context=number", "set number of prior context lines" },
-  { OP_OP_STRING, N_COLOUR, &colour_option,    "color=option",  "matched text color option" },
-  { OP_OP_STRING, N_COLOUR, &colour_option,    "colour=option", "matched text colour option" },
-  { OP_NUMBER,    'C',      &both_context,     "context=number", "set number of context lines, before & after" },
-  { OP_NODATA,    'c',      NULL,              "count",         "print only a count of matching lines per FILE" },
-  { OP_STRING,    'D',      &DEE_option,       "devices=action","how to handle devices, FIFOs, and sockets" },
-  { OP_STRING,    'd',      &dee_option,       "directories=action", "how to handle directories" },
-  { OP_PATLIST,   'e',      NULL,              "regex(p)=pattern", "specify pattern (may be used more than once)" },
-  { OP_NODATA,    'F',      NULL,              "fixed-strings", "patterns are sets of newline-separated strings" },
-  { OP_STRING,    'f',      &pattern_filename, "file=path",     "read patterns from file" },
-  { OP_NODATA,    N_FOFFSETS, NULL,            "file-offsets",  "output file offsets, not text" },
-  { OP_NODATA,    'H',      NULL,              "with-filename", "force the prefixing filename on output" },
-  { OP_NODATA,    'h',      NULL,              "no-filename",   "suppress the prefixing filename on output" },
-  { OP_NODATA,    'i',      NULL,              "ignore-case",   "ignore case distinctions" },
-  { OP_NODATA,    'l',      NULL,              "files-with-matches", "print only FILE names containing matches" },
-  { OP_NODATA,    'L',      NULL,              "files-without-match","print only FILE names not containing matches" },
-  { OP_STRING,    N_LABEL,  &stdin_name,       "label=name",    "set name for standard input" },
-  { OP_NODATA,    N_LBUFFER, NULL,             "line-buffered", "use line buffering" },
-  { OP_NODATA,    N_LOFFSETS, NULL,            "line-offsets",  "output line numbers and offsets, not text" },
-  { OP_STRING,    N_LOCALE, &locale,           "locale=locale", "use the named locale" },
-  { OP_NUMBER,    N_M_LIMIT,&match_limit,      "match-limit=number", "set PCRE match limit option" },
-  { OP_NUMBER,    N_M_LIMIT_REC,&match_limit_recursion, "recursion-limit=number", "set PCRE match recursion limit option" },
-  { OP_NODATA,    'M',      NULL,              "multiline",     "run in multiline mode" },
-  { OP_STRING,    'N',      &newline,          "newline=type",  "set newline type (CR, LF, CRLF, ANYCRLF or ANY)" },
-  { OP_NODATA,    'n',      NULL,              "line-number",   "print line number with output lines" },
-  { OP_OP_NUMBER, 'o',      &only_matching,    "only-matching=n", "show only the part of the line that matched" },
-  { OP_NODATA,    'q',      NULL,              "quiet",         "suppress output, just set return code" },
-  { OP_NODATA,    'r',      NULL,              "recursive",     "recursively scan sub-directories" },
-  { OP_STRING,    N_EXCLUDE,&exclude_pattern,  "exclude=pattern","exclude matching files when recursing" },
-  { OP_STRING,    N_INCLUDE,&include_pattern,  "include=pattern","include matching files when recursing" },
-  { OP_STRING,    N_EXCLUDE_DIR,&exclude_dir_pattern, "exclude-dir=pattern","exclude matching directories when recursing" },
-  { OP_STRING,    N_INCLUDE_DIR,&include_dir_pattern, "include-dir=pattern","include matching directories when recursing" },
+  { OP_NODATA,     N_NULL,   NULL,              "",              "  terminate options" },
+  { OP_NODATA,     N_HELP,   NULL,              "help",          "display this help and exit" },
+  { OP_NUMBER,     'A',      &after_context,    "after-context=number", "set number of following context lines" },
+  { OP_NUMBER,     'B',      &before_context,   "before-context=number", "set number of prior context lines" },
+  { OP_NUMBER,     N_BUFSIZE,&bufthird,         "buffer-size=number", "set processing buffer size parameter" },
+  { OP_OP_STRING,  N_COLOUR, &colour_option,    "color=option",  "matched text color option" },
+  { OP_OP_STRING,  N_COLOUR, &colour_option,    "colour=option", "matched text colour option" },
+  { OP_NUMBER,     'C',      &both_context,     "context=number", "set number of context lines, before & after" },
+  { OP_NODATA,     'c',      NULL,              "count",         "print only a count of matching lines per FILE" },
+  { OP_STRING,     'D',      &DEE_option,       "devices=action","how to handle devices, FIFOs, and sockets" },
+  { OP_STRING,     'd',      &dee_option,       "directories=action", "how to handle directories" },
+  { OP_PATLIST,    'e',      NULL,              "regex(p)=pattern", "specify pattern (may be used more than once)" },
+  { OP_NODATA,     'F',      NULL,              "fixed-strings", "patterns are sets of newline-separated strings" },
+  { OP_STRING,     'f',      &pattern_filename, "file=path",     "read patterns from file" },
+  { OP_NODATA,     N_FOFFSETS, NULL,            "file-offsets",  "output file offsets, not text" },
+  { OP_NODATA,     'H',      NULL,              "with-filename", "force the prefixing filename on output" },
+  { OP_NODATA,     'h',      NULL,              "no-filename",   "suppress the prefixing filename on output" },
+  { OP_NODATA,     'i',      NULL,              "ignore-case",   "ignore case distinctions" },
+#ifdef SUPPORT_PCREGREP_JIT
+  { OP_NODATA,     N_NOJIT,  NULL,              "no-jit",        "do not use just-in-time compiler optimization" },
+#else
+  { OP_NODATA,     N_NOJIT,  NULL,              "no-jit",        "ignored: this pcregrep does not support JIT" },
+#endif
+  { OP_NODATA,     'l',      NULL,              "files-with-matches", "print only FILE names containing matches" },
+  { OP_NODATA,     'L',      NULL,              "files-without-match","print only FILE names not containing matches" },
+  { OP_STRING,     N_LABEL,  &stdin_name,       "label=name",    "set name for standard input" },
+  { OP_NODATA,     N_LBUFFER, NULL,             "line-buffered", "use line buffering" },
+  { OP_NODATA,     N_LOFFSETS, NULL,            "line-offsets",  "output line numbers and offsets, not text" },
+  { OP_STRING,     N_LOCALE, &locale,           "locale=locale", "use the named locale" },
+  { OP_LONGNUMBER, N_M_LIMIT, &match_limit,     "match-limit=number", "set PCRE match limit option" },
+  { OP_LONGNUMBER, N_M_LIMIT_REC, &match_limit_recursion, "recursion-limit=number", "set PCRE match recursion limit option" },
+  { OP_NODATA,     'M',      NULL,              "multiline",     "run in multiline mode" },
+  { OP_STRING,     'N',      &newline,          "newline=type",  "set newline type (CR, LF, CRLF, ANYCRLF or ANY)" },
+  { OP_NODATA,     'n',      NULL,              "line-number",   "print line number with output lines" },
+  { OP_OP_NUMBER,  'o',      &only_matching,    "only-matching=n", "show only the part of the line that matched" },
+  { OP_NODATA,     'q',      NULL,              "quiet",         "suppress output, just set return code" },
+  { OP_NODATA,     'r',      NULL,              "recursive",     "recursively scan sub-directories" },
+  { OP_STRING,     N_EXCLUDE,&exclude_pattern,  "exclude=pattern","exclude matching files when recursing" },
+  { OP_STRING,     N_INCLUDE,&include_pattern,  "include=pattern","include matching files when recursing" },
+  { OP_STRING,     N_EXCLUDE_DIR,&exclude_dir_pattern, "exclude-dir=pattern","exclude matching directories when recursing" },
+  { OP_STRING,     N_INCLUDE_DIR,&include_dir_pattern, "include-dir=pattern","include matching directories when recursing" },
 
   /* These two were accidentally implemented with underscores instead of
   hyphens in the option names. As this was not discovered for several releases,
@@ -295,6 +312,31 @@ const char utf8_table4[] = {
 
 
 
+/*************************************************
+*         Exit from the program                  *
+*************************************************/
+
+/* If there has been a resource error, give a suitable message.
+
+Argument:  the return code
+Returns:   does not return
+*/
+
+static void
+pcregrep_exit(int rc)
+{
+if (resource_error)
+  {
+  fprintf(stderr, "pcregrep: Error %d, %d or %d means that a resource limit "
+    "was exceeded.\n", PCRE_ERROR_MATCHLIMIT, PCRE_ERROR_RECURSIONLIMIT,
+    PCRE_ERROR_JIT_STACKLIMIT);
+  fprintf(stderr, "pcregrep: Check your regex for nested unlimited loops.\n");
+  }
+
+exit(rc);
+}
+
+
 /*************************************************
 *            OS-specific functions               *
 *************************************************/
@@ -564,31 +606,6 @@ return sys_errlist[n];
 
 
 
-/*************************************************
-*         Exit from the program                  *
-*************************************************/
-
-/* If there has been a resource error, give a suitable message.
-
-Argument:  the return code
-Returns:   does not return
-*/
-
-static void
-pcregrep_exit(int rc)
-{
-if (resource_error)
-  {
-  fprintf(stderr, "pcregrep: Error %d or %d means that a resource limit "
-    "was exceeded.\n", PCRE_ERROR_MATCHLIMIT, PCRE_ERROR_RECURSIONLIMIT);
-  fprintf(stderr, "pcregrep: Check your regex for nested unlimited loops.\n");
-  }
-
-exit(rc);
-}
-
-
-
 /*************************************************
 *            Read one line of input              *
 *************************************************/
@@ -635,7 +652,8 @@ Arguments:
   endptr    end of available data
   lenptr    where to put the length of the eol sequence
 
-Returns:    pointer to the last byte of the line
+Returns:    pointer after the last byte of the line,
+            including the newline byte(s)
 */
 
 static char *
@@ -936,10 +954,11 @@ is used multiple times for the same subject when colouring is enabled, in order
 to find all possible matches.
 
 Arguments:
-  matchptr    the start of the subject
-  length      the length of the subject to match
-  offsets     the offets vector to fill in
-  mrc         address of where to put the result of pcre_exec()
+  matchptr     the start of the subject
+  length       the length of the subject to match
+  startoffset  where to start matching
+  offsets      the offets vector to fill in
+  mrc          address of where to put the result of pcre_exec()
 
 Returns:      TRUE if there was a match
               FALSE if there was no match
@@ -947,7 +966,8 @@ Returns:      TRUE if there was a match
 */
 
 static BOOL
-match_patterns(char *matchptr, size_t length, int *offsets, int *mrc)
+match_patterns(char *matchptr, size_t length, int startoffset, int *offsets,
+  int *mrc)
 {
 int i;
 size_t slen = length;
@@ -959,8 +979,8 @@ if (slen > 200)
   }
 for (i = 0; i < pattern_count; i++)
   {
-  *mrc = pcre_exec(pattern_list[i], hints_list[i], matchptr, (int)length, 0,
-    PCRE_NOTEMPTY, offsets, OFFSET_SIZE);
+  *mrc = pcre_exec(pattern_list[i], hints_list[i], matchptr, (int)length,
+    startoffset, PCRE_NOTEMPTY, offsets, OFFSET_SIZE);
   if (*mrc >= 0) return TRUE;
   if (*mrc == PCRE_ERROR_NOMATCH) continue;
   fprintf(stderr, "pcregrep: pcre_exec() gave error %d while matching ", *mrc);
@@ -968,7 +988,8 @@ for (i = 0; i < pattern_count; i++)
   fprintf(stderr, "%s", msg);
   FWRITE(matchptr, 1, slen, stderr);   /* In case binary zero included */
   fprintf(stderr, "\n\n");
-  if (*mrc == PCRE_ERROR_MATCHLIMIT || *mrc == PCRE_ERROR_RECURSIONLIMIT)
+  if (*mrc == PCRE_ERROR_MATCHLIMIT || *mrc == PCRE_ERROR_RECURSIONLIMIT ||
+      *mrc == PCRE_ERROR_JIT_STACKLIMIT)
     resource_error = TRUE;
   if (error_count++ > 20)
     {
@@ -988,7 +1009,7 @@ return FALSE;  /* No match, no errors */
 *************************************************/
 
 /* This is called from grep_or_recurse() below. It uses a buffer that is three
-times the value of MBUFTHIRD. The matching point is never allowed to stray into
+times the value of bufthird. The matching point is never allowed to stray into
 the top third of the buffer, thus keeping more of the file available for
 context printing or for multiline scanning. For large files, the pointer will
 be in the middle third most of the time, so the bottom third is available for
@@ -999,17 +1020,19 @@ Arguments:
                the gzFile pointer when reading is via libz
                the BZFILE pointer when reading is via libbz2
   frtype       FR_PLAIN, FR_LIBZ, or FR_LIBBZ2
+  filename     the file name or NULL (for errors)
   printname    the file name if it is to be printed for each match
                or NULL if the file name is not to be printed
                it cannot be NULL if filenames[_nomatch]_only is set
 
 Returns:       0 if there was at least one match
                1 otherwise (no matches)
-               2 if there is a read error on a .bz2 file
+               2 if an overlong line is encountered
+               3 if there is a read error on a .bz2 file
 */
 
 static int
-pcregrep(void *handle, int frtype, char *printname)
+pcregrep(void *handle, int frtype, char *filename, char *printname)
 {
 int rc = 1;
 int linenumber = 1;
@@ -1018,8 +1041,7 @@ int count = 0;
 int filepos = 0;
 int offsets[OFFSET_SIZE];
 char *lastmatchrestart = NULL;
-char buffer[3*MBUFTHIRD];
-char *ptr = buffer;
+char *ptr = main_buffer;
 char *endptr;
 size_t bufflength;
 BOOL endhyphenpending = FALSE;
@@ -1044,7 +1066,7 @@ fail. */
 if (frtype == FR_LIBZ)
   {
   ingz = (gzFile)handle;
-  bufflength = gzread (ingz, buffer, 3*MBUFTHIRD);
+  bufflength = gzread (ingz, main_buffer, bufsize);
   }
 else
 #endif
@@ -1053,7 +1075,7 @@ else
 if (frtype == FR_LIBBZ2)
   {
   inbz2 = (BZFILE *)handle;
-  bufflength = BZ2_bzread(inbz2, buffer, 3*MBUFTHIRD);
+  bufflength = BZ2_bzread(inbz2, main_buffer, bufsize);
   if ((int)bufflength < 0) return 2;   /* Gotcha: bufflength is size_t; */
   }                                    /* without the cast it is unsigned. */
 else
@@ -1063,11 +1085,11 @@ else
   in = (FILE *)handle;
   if (is_file_tty(in)) input_line_buffered = TRUE;
   bufflength = input_line_buffered?
-    read_one_line(buffer, 3*MBUFTHIRD, in) :
-    fread(buffer, 1, 3*MBUFTHIRD, in);
+    read_one_line(main_buffer, bufsize, in) :
+    fread(main_buffer, 1, bufsize, in);
   }
 
-endptr = buffer + bufflength;
+endptr = main_buffer + bufflength;
 
 /* Loop while the current pointer is not at the end of the file. For large
 files, endptr will be at the end of the buffer when we are in the middle of the
@@ -1078,6 +1100,7 @@ while (ptr < endptr)
   {
   int endlinelength;
   int mrc = 0;
+  int startoffset = 0;
   BOOL match;
   char *matchptr = ptr;
   char *t = ptr;
@@ -1095,6 +1118,20 @@ while (ptr < endptr)
   linelength = t - ptr - endlinelength;
   length = multiline? (size_t)(endptr - ptr) : linelength;
 
+  /* Check to see if the line we are looking at extends right to the very end
+  of the buffer without a line terminator. This means the line is too long to
+  handle. */
+
+  if (endlinelength == 0 && t == main_buffer + bufsize)
+    {
+    fprintf(stderr, "pcregrep: line %d%s%s is too long for the internal buffer\n"
+                    "pcregrep: check the --buffer-size option\n",
+                    linenumber,
+                    (filename == NULL)? "" : " of file ",
+                    (filename == NULL)? "" : filename);
+    return 2;
+    }
+
   /* Extra processing for Jeffrey Friedl's debugging. */
 
 #ifdef JFRIEDL_DEBUG
@@ -1154,7 +1191,7 @@ while (ptr < endptr)
   than NOMATCH. This code is in a subroutine so that it can be re-used for
   finding subsequent matches when colouring matched lines. */
 
-  match = match_patterns(matchptr, length, offsets, &mrc);
+  match = match_patterns(matchptr, length, startoffset, offsets, &mrc);
 
   /* If it's a match or a not-match (as required), do what's wanted. */
 
@@ -1187,10 +1224,10 @@ while (ptr < endptr)
     captured portion of it, as long as this string is not empty, and the
     --file-offsets and --line-offsets options output offsets for the matching
     substring (they both force --only-matching = 0). None of these options
-    prints any context. Afterwards, adjust the start and length, and then jump
-    back to look for further matches in the same line. If we are in invert
-    mode, however, nothing is printed and we do not restart - this could still
-    be useful because the return code is set. */
+    prints any context. Afterwards, adjust the start and then jump back to look
+    for further matches in the same line. If we are in invert mode, however,
+    nothing is printed and we do not restart - this could still be useful
+    because the return code is set. */
 
     else if (only_matching >= 0)
       {
@@ -1217,11 +1254,10 @@ while (ptr < endptr)
             }
           }
         else if (printname != NULL || number) fprintf(stdout, "\n");
-        matchptr += offsets[1];
-        length -= offsets[1];
         match = FALSE;
         if (line_buffered) fflush(stdout);
-        rc = 0;    /* Had some success */
+        rc = 0;                      /* Had some success */
+        startoffset = offsets[1];    /* Restart after the match */
         goto ONLY_MATCHING_RESTART;
         }
       }
@@ -1280,11 +1316,11 @@ while (ptr < endptr)
         int linecount = 0;
         char *p = ptr;
 
-        while (p > buffer && (lastmatchnumber == 0 || p > lastmatchrestart) &&
+        while (p > main_buffer && (lastmatchnumber == 0 || p > lastmatchrestart) &&
                linecount < before_context)
           {
           linecount++;
-          p = previous_line(p, buffer);
+          p = previous_line(p, main_buffer);
           }
 
         if (lastmatchnumber > 0 && p > lastmatchrestart && !hyphenprinted)
@@ -1317,22 +1353,16 @@ while (ptr < endptr)
       (invert not set). Because the PCRE_FIRSTLINE option is set, the start of
       the match will always be before the first newline sequence. */
 
-      if (multiline)
+      if (multiline & !invert)
         {
-        int ellength;
-        char *endmatch = ptr;
-        if (!invert)
+        char *endmatch = ptr + offsets[1];
+        t = ptr;
+        while (t < endmatch)
           {
-          endmatch += offsets[1];
-          t = ptr;
-          while (t < endmatch)
-            {
-            t = end_of_line(t, endptr, &ellength);
-            if (t <= endmatch) linenumber++; else break;
-            }
+          t = end_of_line(t, endptr, &endlinelength);
+          if (t < endmatch) linenumber++; else break;
           }
-        endmatch = end_of_line(endmatch, endptr, &ellength);
-        linelength = endmatch - ptr - ellength;
+        linelength = t - ptr - endlinelength;
         }
 
       /*** NOTE: Use only fwrite() to output the data line, so that binary
@@ -1355,28 +1385,33 @@ while (ptr < endptr)
 #endif
 
       /* We have to split the line(s) up if colouring, and search for further
-      matches. */
+      matches, but not of course if the line is a non-match. */
 
-      if (do_colour)
+      if (do_colour && !invert)
         {
-        int last_offset = 0;
+        int plength;
         FWRITE(ptr, 1, offsets[0], stdout);
         fprintf(stdout, "%c[%sm", 0x1b, colour_string);
         FWRITE(ptr + offsets[0], 1, offsets[1] - offsets[0], stdout);
         fprintf(stdout, "%c[00m", 0x1b);
         for (;;)
           {
-          last_offset += offsets[1];
-          matchptr += offsets[1];
-          length -= offsets[1];
-          if (!match_patterns(matchptr, length, offsets, &mrc)) break;
-          FWRITE(matchptr, 1, offsets[0], stdout);
+          startoffset = offsets[1];
+          if (startoffset >= (int)linelength + endlinelength ||
+              !match_patterns(matchptr, length, startoffset, offsets, &mrc))
+            break;
+          FWRITE(matchptr + startoffset, 1, offsets[0] - startoffset, stdout);
           fprintf(stdout, "%c[%sm", 0x1b, colour_string);
           FWRITE(matchptr + offsets[0], 1, offsets[1] - offsets[0], stdout);
           fprintf(stdout, "%c[00m", 0x1b);
           }
-        FWRITE(ptr + last_offset, 1,
-          (linelength + endlinelength) - last_offset, stdout);
+
+        /* In multiline mode, we may have already printed the complete line
+        and its line-ending characters (if they matched the pattern), so there
+        may be no more to print. */
+
+        plength = (linelength + endlinelength) - startoffset;
+        if (plength > 0) FWRITE(ptr + startoffset, 1, plength, stdout);
         }
 
       /* Not colouring; no need to search for further matches */
@@ -1425,9 +1460,9 @@ while (ptr < endptr)
   /* If input is line buffered, and the buffer is not yet full, read another
   line and add it into the buffer. */
 
-  if (input_line_buffered && bufflength < sizeof(buffer))
+  if (input_line_buffered && bufflength < (size_t)bufsize)
     {
-    int add = read_one_line(ptr, sizeof(buffer) - (ptr - buffer), in);
+    int add = read_one_line(ptr, bufsize - (ptr - main_buffer), in);
     bufflength += add;
     endptr += add;
     }
@@ -1437,11 +1472,11 @@ while (ptr < endptr)
   1/3 and refill it. Before we do this, if some unprinted "after" lines are
   about to be lost, print them. */
 
-  if (bufflength >= sizeof(buffer) && ptr > buffer + 2*MBUFTHIRD)
+  if (bufflength >= (size_t)bufsize && ptr > main_buffer + 2*bufthird)
     {
     if (after_context > 0 &&
         lastmatchnumber > 0 &&
-        lastmatchrestart < buffer + MBUFTHIRD)
+        lastmatchrestart < main_buffer + bufthird)
       {
       do_after_lines(lastmatchnumber, lastmatchrestart, endptr, printname);
       lastmatchnumber = 0;
@@ -1449,32 +1484,32 @@ while (ptr < endptr)
 
     /* Now do the shuffle */
 
-    memmove(buffer, buffer + MBUFTHIRD, 2*MBUFTHIRD);
-    ptr -= MBUFTHIRD;
+    memmove(main_buffer, main_buffer + bufthird, 2*bufthird);
+    ptr -= bufthird;
 
 #ifdef SUPPORT_LIBZ
     if (frtype == FR_LIBZ)
-      bufflength = 2*MBUFTHIRD +
-        gzread (ingz, buffer + 2*MBUFTHIRD, MBUFTHIRD);
+      bufflength = 2*bufthird +
+        gzread (ingz, main_buffer + 2*bufthird, bufthird);
     else
 #endif
 
 #ifdef SUPPORT_LIBBZ2
     if (frtype == FR_LIBBZ2)
-      bufflength = 2*MBUFTHIRD +
-        BZ2_bzread(inbz2, buffer + 2*MBUFTHIRD, MBUFTHIRD);
+      bufflength = 2*bufthird +
+        BZ2_bzread(inbz2, main_buffer + 2*bufthird, bufthird);
     else
 #endif
 
-    bufflength = 2*MBUFTHIRD +
+    bufflength = 2*bufthird +
       (input_line_buffered?
-       read_one_line(buffer + 2*MBUFTHIRD, MBUFTHIRD, in) :
-       fread(buffer + 2*MBUFTHIRD, 1, MBUFTHIRD, in));
-    endptr = buffer + bufflength;
+       read_one_line(main_buffer + 2*bufthird, bufthird, in) :
+       fread(main_buffer + 2*bufthird, 1, bufthird, in));
+    endptr = main_buffer + bufflength;
 
     /* Adjust any last match point */
 
-    if (lastmatchnumber > 0) lastmatchrestart -= MBUFTHIRD;
+    if (lastmatchnumber > 0) lastmatchrestart -= bufthird;
     }
   }     /* Loop through the whole file */
 
@@ -1554,7 +1589,7 @@ BZFILE *inbz2 = NULL;
 
 if (strcmp(pathname, "-") == 0)
   {
-  return pcregrep(stdin, FR_PLAIN,
+  return pcregrep(stdin, FR_PLAIN, stdin_name,
     (filenames > FN_DEFAULT || (filenames == FN_DEFAULT && !only_one_at_top))?
       stdin_name : NULL);
   }
@@ -1686,7 +1721,7 @@ if (handle == NULL)
 
 /* Now grep the file */
 
-rc = pcregrep(handle, frtype, (filenames > FN_DEFAULT ||
+rc = pcregrep(handle, frtype, pathname, (filenames > FN_DEFAULT ||
   (filenames == FN_DEFAULT && !only_one_at_top))? pathname : NULL);
 
 /* Close in an appropriate manner. */
@@ -1697,14 +1732,14 @@ if (frtype == FR_LIBZ)
 else
 #endif
 
-/* If it is a .bz2 file and the result is 2, it means that the first attempt to
+/* If it is a .bz2 file and the result is 3, it means that the first attempt to
 read failed. If the error indicates that the file isn't in fact bzipped, try
 again as a normal file. */
 
 #ifdef SUPPORT_LIBBZ2
 if (frtype == FR_LIBBZ2)
   {
-  if (rc == 2)
+  if (rc == 3)
     {
     int errnum;
     const char *err = BZ2_bzerror(inbz2, &errnum);
@@ -1716,6 +1751,7 @@ if (frtype == FR_LIBBZ2)
     else if (!silent)
       fprintf(stderr, "pcregrep: Failed to read %s using bzlib: %s\n",
         pathname, err);
+    rc = 2;    /* The normal "something went wrong" code */
     }
   BZ2_bzclose(inbz2);
   }
@@ -1808,9 +1844,12 @@ for (op = optionlist; op->one_char != 0; op++)
   printf("%.*s%s\n", n, "                     ", op->help_text);
   }
 
-printf("\nWhen reading patterns from a file instead of using a command line option,\n");
+printf("\nNumbers may be followed by K or M, e.g. --buffer-size=100K.\n");
+printf("The default value for --buffer-size is %d.\n", PCREGREP_BUFSIZE);
+printf("When reading patterns from a file instead of using a command line option,\n");
 printf("trailing white space is removed and blank lines are ignored.\n");
-printf("There is a maximum of %d patterns.\n", MAX_PATTERN_COUNT);
+printf("There is a maximum of %d patterns, each of maximum size %d bytes.\n",
+  MAX_PATTERN_COUNT, PATBUFSIZE);
 
 printf("\nWith no FILEs, read standard input. If fewer than two FILEs given, assume -h.\n");
 printf("Exit status is 0 if any matches, 1 if no matches, and 2 if trouble.\n");
@@ -1830,8 +1869,9 @@ switch(letter)
   {
   case N_FOFFSETS: file_offsets = TRUE; break;
   case N_HELP: help(); pcregrep_exit(0);
-  case N_LOFFSETS: line_offsets = number = TRUE; break;
   case N_LBUFFER: line_buffered = TRUE; break;
+  case N_LOFFSETS: line_offsets = number = TRUE; break;
+  case N_NOJIT: study_options &= ~PCRE_STUDY_JIT_COMPILE; break;
   case 'c': count_only = TRUE; break;
   case 'F': process_options |= PO_FIXED_STRINGS; break;
   case 'H': filenames = FN_FORCE; break;
@@ -1912,7 +1952,7 @@ Returns:         TRUE on success, FALSE after an error
 static BOOL
 compile_single_pattern(char *pattern, int options, char *filename, int count)
 {
-char buffer[MBUFTHIRD + 16];
+char buffer[PATBUFSIZE];
 const char *error;
 int errptr;
 
@@ -1923,7 +1963,7 @@ if (pattern_count >= MAX_PATTERN_COUNT)
   return FALSE;
   }
 
-sprintf(buffer, "%s%.*s%s", prefix[process_options], MBUFTHIRD, pattern,
+sprintf(buffer, "%s%.*s%s", prefix[process_options], bufthird, pattern,
   suffix[process_options]);
 pattern_list[pattern_count] =
   pcre_compile(buffer, options, &error, &errptr, pcretables);
@@ -1982,7 +2022,7 @@ compile_pattern(char *pattern, int options, char *filename, int count)
 if ((process_options & PO_FIXED_STRINGS) != 0)
   {
   char *eop = pattern + strlen(pattern);
-  char buffer[MBUFTHIRD];
+  char buffer[PATBUFSIZE];
   for(;;)
     {
     int ellength;
@@ -2020,6 +2060,10 @@ char *patterns[MAX_PATTERN_COUNT];
 const char *locale_from = "--locale";
 const char *error;
 
+#ifdef SUPPORT_PCREGREP_JIT
+pcre_jit_stack *jit_stack = NULL;
+#endif
+
 /* Set the default line ending value from the default in the PCRE library;
 "lf", "cr", "crlf", and "any" are supported. Anything else is treated as "lf".
 Note that the return values from pcre_config(), though derived from the ASCII
@@ -2278,7 +2322,8 @@ for (i = 1; i < argc; i++)
 
   /* Otherwise, deal with single string or numeric data values. */
 
-  else if (op->type != OP_NUMBER && op->type != OP_OP_NUMBER)
+  else if (op->type != OP_NUMBER && op->type != OP_LONGNUMBER &&
+           op->type != OP_OP_NUMBER)
     {
     *((char **)op->dataptr) = option_data;
     }
@@ -2293,6 +2338,16 @@ for (i = 1; i < argc; i++)
     while (*endptr != 0 && isspace((unsigned char)(*endptr))) endptr++;
     while (isdigit((unsigned char)(*endptr)))
       n = n * 10 + (int)(*endptr++ - '0');
+    if (toupper(*endptr) == 'K')
+      {
+      n *= 1024;
+      endptr++;
+      }
+    else if (toupper(*endptr) == 'M')
+      {
+      n *= 1024*1024;
+      endptr++;
+      }
     if (*endptr != 0)
       {
       if (longop)
@@ -2308,7 +2363,10 @@ for (i = 1; i < argc; i++)
           option_data, op->one_char);
       pcregrep_exit(usage(2));
       }
-    *((int *)op->dataptr) = n;
+    if (op->type == OP_LONGNUMBER)
+        *((unsigned long int *)op->dataptr) = n;
+    else
+        *((int *)op->dataptr) = n;
     }
   }
 
@@ -2456,12 +2514,14 @@ if (jfriedl_XT != 0 || jfriedl_XR != 0)
   }
 #endif
 
-/* Get memory to store the pattern and hints lists. */
+/* Get memory for the main buffer, and to store the pattern and hints lists. */
 
+bufsize = 3*bufthird;
+main_buffer = (char *)malloc(bufsize);
 pattern_list = (pcre **)malloc(MAX_PATTERN_COUNT * sizeof(pcre *));
 hints_list = (pcre_extra **)malloc(MAX_PATTERN_COUNT * sizeof(pcre_extra *));
 
-if (pattern_list == NULL || hints_list == NULL)
+if (main_buffer == NULL || pattern_list == NULL || hints_list == NULL)
   {
   fprintf(stderr, "pcregrep: malloc failed\n");
   goto EXIT2;
@@ -2493,7 +2553,7 @@ if (pattern_filename != NULL)
   int linenumber = 0;
   FILE *f;
   char *filename;
-  char buffer[MBUFTHIRD];
+  char buffer[PATBUFSIZE];
 
   if (strcmp(pattern_filename, "-") == 0)
     {
@@ -2512,7 +2572,7 @@ if (pattern_filename != NULL)
     filename = pattern_filename;
     }
 
-  while (fgets(buffer, MBUFTHIRD, f) != NULL)
+  while (fgets(buffer, PATBUFSIZE, f) != NULL)
     {
     char *s = buffer + (int)strlen(buffer);
     while (s > buffer && isspace((unsigned char)(s[-1]))) s--;
@@ -2526,11 +2586,17 @@ if (pattern_filename != NULL)
   if (f != stdin) fclose(f);
   }
 
-/* Study the regular expressions, as we will be running them many times */
+/* Study the regular expressions, as we will be running them many times. Unless
+JIT has been explicitly disabled, arrange a stack for it to use. */
+
+#ifdef SUPPORT_PCREGREP_JIT
+if ((study_options & PCRE_STUDY_JIT_COMPILE) != 0)
+  jit_stack = pcre_jit_stack_alloc(32*1024, 1024*1024);
+#endif
 
 for (j = 0; j < pattern_count; j++)
   {
-  hints_list[j] = pcre_study(pattern_list[j], 0, &error);
+  hints_list[j] = pcre_study(pattern_list[j], study_options, &error);
   if (error != NULL)
     {
     char s[16];
@@ -2539,6 +2605,10 @@ for (j = 0; j < pattern_count; j++)
     goto EXIT2;
     }
   hint_count++;
+#ifdef SUPPORT_PCREGREP_JIT
+  if (jit_stack != NULL && hints_list[j] != NULL)
+    pcre_assign_jit_stack(hints_list[j], NULL, jit_stack);
+#endif
   }
 
 /* If --match-limit or --recursion-limit was set, put the value(s) into the
@@ -2624,7 +2694,8 @@ if (include_dir_pattern != NULL)
 
 if (i >= argc)
   {
-  rc = pcregrep(stdin, FR_PLAIN, (filenames > FN_DEFAULT)? stdin_name : NULL);
+  rc = pcregrep(stdin, FR_PLAIN, stdin_name,
+    (filenames > FN_DEFAULT)? stdin_name : NULL);
   goto EXIT;
   }
 
@@ -2644,6 +2715,10 @@ for (; i < argc; i++)
   }
 
 EXIT:
+#ifdef SUPPORT_PCREGREP_JIT
+if (jit_stack != NULL) pcre_jit_stack_free(jit_stack);
+#endif
+if (main_buffer != NULL) free(main_buffer);
 if (pattern_list != NULL)
   {
   for (i = 0; i < pattern_count; i++) free(pattern_list[i]);
@@ -2653,7 +2728,7 @@ if (hints_list != NULL)
   {
   for (i = 0; i < hint_count; i++)
     {
-    if (hints_list[i] != NULL) free(hints_list[i]);
+    if (hints_list[i] != NULL) pcre_free_study(hints_list[i]);
     }
   free(hints_list);
   }
